import os
import configparser


DEFAULT_REF_CONFIG = 'refs.cfg'

CONFIG_TEMPL = """
    [info]
    species =
    source =
    version =

    [refs]
    anno_db =
    ref_file =
    mir_ref =
    mir_target =
    other_transcripts =
    AGO_data =
    RBP_data =
    RBP_target =
"""


class RefConfig:
    def __init__(self):
        self.config = configparser.ConfigParser()
        self.config.optionxform = str
        self.config.read_string(CONFIG_TEMPL)

    def read(self, cfg_file):
        self.config.read(cfg_file)

    def write(self, ref_dir, cfg_file_name=DEFAULT_REF_CONFIG):
        cfg_file = os.path.join(ref_dir, cfg_file_name)
        with open(cfg_file, 'w') as config_file:
            self.config.write(config_file)

    def __getitem__(self, key):
        return self.config[key]


def get_refs(ref_dir):
    cfg_file = os.path.join(ref_dir, DEFAULT_REF_CONFIG)

    configObj = RefConfig()
    configObj.read(cfg_file)
    config = configObj.config

    anno_db = prepend_dirname_to_file(ref_dir, config['refs']['anno_db'])
    ref_file = prepend_dirname_to_file(ref_dir, config['refs']['ref_file'])
    mir_ref = prepend_dirname_to_file(ref_dir, config['refs']['mir_ref'])
    mir_target = prepend_dirname_to_file(ref_dir, config['refs']['mir_target'])
    other_transcripts = prepend_dirname_to_file(ref_dir, config['refs']['other_transcripts'])
    AGO_data = prepend_dirname_to_file(ref_dir, config['refs']['AGO_data'])
    RBP_data = None #prepend_dirname_to_file(ref_dir, config['refs']['RBP_data'])
    RBP_target = None #prepend_dirname_to_file(ref_dir, config['refs']['RBP_target'])

    return (anno_db,
            ref_file,
            mir_ref,
            mir_target,
            other_transcripts,
            AGO_data,
            RBP_data,
            RBP_target)


def prepend_dirname_to_file(ref_dir, filename):
    if filename == '':
        return None
    else:
        return os.path.join(ref_dir, filename)



import networkx as nx
from circmimi.network.utils import read_table
from circmimi.network.formats.xgmml import write_xgmml


class CyNetwork:
    def __init__(self):
        self.graph = nx.DiGraph()

    def load_data(self, file_, k1, k2, k3, header=True):
        for data in read_table(file_, header=header):
            circRNA = data[k1 - 1]
            mediator = data[k2 - 1]
            target_gene = data[k3 - 1]

            self.graph.add_node(
                circRNA,
                data_type="circRNA"
            )

            self.graph.add_node(
                mediator,
                data_type="mediator"
            )

            self.graph.add_node(
                target_gene,
                data_type='target_gene'
            )

            self.graph.add_edge(
                circRNA,
                mediator,
                data_type="circRNA-mediator"
            )

            self.graph.add_edge(
                mediator,
                target_gene,
                data_type="mediator-target_gene"
            )

    def update_node_attr(self, node, key, val):
        self.graph.nodes[node][key] = val

    def update_edge_attr(self, edge, key, val):
        self.graph.edges[edge][key] = val

    def apply_layout(self, layout):
        layout.apply(self.graph)

    def apply_style(self, style):
        style.apply(self.graph)

    def to_xgmml(self, file_):
        write_xgmml(self.graph, file_)


class Layout:
    def __init__(self):
        pass

    def apply(self, graph):
        pass


class Style:
    def __init__(self):
        self.circRNAs_style = {'type': 'circle', 'fill': '#F292AF'}
        self.mediator_style = {'type': 'ROUNDED_RECTANGLE', 'fill': '#8BE26E'}
        self.target_gene_style = {'type': 'diamond', 'fill': '#6D9AF4'}

        self.circRNAs_mediator_style = {'fill': '#FFC1C1'}
        self.mediator_taget_gene_style = {'fill': '#9BDDA2'}

    def apply(self, graph):
        for node in graph.nodes:
            node_type = graph.nodes[node]['data_type']
            if node_type == 'circRNA':
                graph.nodes[node]['graphics'] = self.circRNAs_style
            elif node_type == 'mediator':
                graph.nodes[node]['graphics'] = self.mediator_style
            elif node_type == 'target_gene':
                graph.nodes[node]['graphics'] = self.target_gene_style

        for edge in graph.edges:
            edge_type = graph.edges[edge]['data_type']
            if edge_type == 'circRNA-mediator':
                graph.edges[edge]['graphics'] = self.circRNAs_mediator_style
            elif edge_type == 'mediator-target_gene':
                graph.edges[edge]['graphics'] = self.mediator_taget_gene_style



import os.path


def add_prefix(filename, prefix, auto_dot=False):
    if auto_dot:
        prefix_dir, prefix_base = os.path.split(prefix)

        filename = "{}.{}".format(prefix_base, filename)
        filename = filename.lstrip('.')
        filename = os.path.join(prefix_dir, filename)
    else:
        filename = prefix + filename

    return filename



import re
import numpy as np
from scipy.stats import hypergeom
from statsmodels.stats import multitest
from operator import itemgetter


def _count_miRNAs(mir_ref_file):
    with open(mir_ref_file) as f_in:
        mir_ref = f_in.read()

    num_miRNAs = len(re.findall(r'^>', mir_ref, flags=re.M))
    return num_miRNAs


def _calc_hypergeom_pvalue(s, rvs_dict):
    k = s['#miRNAs_share']
    M = s['#miRNAs']
    n = s['#miRNAs_target_gene']
    N = s['#miRNAs_circRNAs']

    rv = rvs_dict[(M, n, N)]
    x = np.arange(k, min(n, N))
    pvalue = rv.pmf(x).sum()

    return pvalue


def do_the_calculation_for_hypergeom_pvalue(circ_target_df):
    all_M_n_N = circ_target_df[[
        '#miRNAs',
        '#miRNAs_target_gene',
        '#miRNAs_circRNAs'
    ]].drop_duplicates().values

    all_M_n_N_tuple = [tuple(row) for row in all_M_n_N]

    rvs = np.apply_along_axis(lambda row: hypergeom(*row), 1, all_M_n_N)
    rvs_dict = dict(zip(all_M_n_N_tuple, rvs))

    p_values = circ_target_df.apply(_calc_hypergeom_pvalue, rvs_dict=rvs_dict, axis=1)

    circ_target_df_with_pv = circ_target_df.assign(p_value=p_values)

    return circ_target_df_with_pv


def do_the_hypergeometric_test(circ_mi_target_df, mir_ref_file, mir_target_db):
    num_miRNAs = _count_miRNAs(mir_ref_file)
    num_miRNAs__circRNA = circ_mi_target_df[['circ_id', 'mirna']].drop_duplicates().groupby('circ_id').agg('count')
    num_miRNAs__circRNA_target_gene = circ_mi_target_df.groupby(['circ_id', 'target_gene']).agg('count')
    num_miRNAs__target_gene = mir_target_db[['target_gene', 'mirna']].drop_duplicates().groupby('target_gene').agg('count')

    circ_target_df = num_miRNAs__circRNA_target_gene.reset_index().rename(
        {
            'mirna': '#miRNAs_share'
        },
        axis=1
    ).merge(
        num_miRNAs__circRNA,
        on='circ_id',
        how='left'
    ).rename(
        {
            'mirna': '#miRNAs_circRNAs'
        },
        axis=1
    ).assign(
        **{
            '#miRNAs': num_miRNAs
        }
    ).merge(
        num_miRNAs__target_gene,
        on='target_gene',
        how='left'
    ).rename(
        {
            'mirna': '#miRNAs_target_gene'
        },
        axis=1
    )

    # do the test
    circ_target_df_with_pv = do_the_calculation_for_hypergeom_pvalue(circ_target_df)

    # adjusting p-values
    bonferroni_corrected_p_values = circ_target_df_with_pv.groupby('circ_id')['p_value'].transform(
        lambda pvs: multitest.multipletests(pvs, method='bonferroni')[1]
    )
    bh_corrected_p_values = circ_target_df_with_pv.groupby('circ_id')['p_value'].transform(
        lambda pvs: multitest.multipletests(pvs, method='fdr_bh')[1]
    )

    circ_target_df_with_pv = circ_target_df_with_pv.assign(
        bh_corrected_p_value=bh_corrected_p_values,
        bonferroni_corrected_p_values=bonferroni_corrected_p_values,
    )

    return circ_target_df_with_pv



import pandas as pd
import io
import subprocess as sp
import tempfile as tp
from collections import namedtuple
from operator import itemgetter


class Bed:
    BED_TITLE = (
        'chr',
        'start',
        'end',
        'name',
        'score',
        'strand',
        'thickStart',
        'thickEnd',
        'itemRGB',
        'blockCount',
        'blockSizes',
        'blockStarts'
    )
    _region = namedtuple('Region', ('start', 'end'))

    def __init__(self, regions, name='.', union=False):
        self.regions = regions
        self.name = name
        self.union = union

        if not self.union:
            self.regions = self.reverse_regions_if_minus_strand(self.regions)
        else:
            self.regions = self.get_union_regions(self.regions)

        self._parse_regions(self.regions)

    @staticmethod
    def reverse_regions_if_minus_strand(regions):
        strand = regions[0][3]
        if strand == '-':
            regions = list(reversed(regions))

        return regions

    def _parse_regions(self, regions):
        self._parse_region(regions[0])
        for region in regions[1:]:
            self._add_block(region)

    def _parse_region(self, region):
        self.chrom = region[0]
        self.start = int(region[1]) - 1
        self.end = int(region[2])
        self.strand = region[3]

        self.block_count = 1
        self.block_sizes = [self.end - self.start]
        self.block_starts = [0]

    def _add_block(self, region):
        other = Bed([region])

        assert self.strand == other.strand

        self.block_count += other.block_count
        self.block_sizes += other.block_sizes

        all_starts = [
            start + self.start for start in self.block_starts
        ] + [other.start]

        if other.start < self.start:
            self.start = other.start

        self.block_starts = [start - self.start for start in all_starts]

        if other.end > self.end:
            self.end = other.end

    def get_data(self, all_fields=False):
        fields = [
            self.chrom,
            self.start,
            self.end,
            self.name,
            '.',
            self.strand
        ]

        if all_fields:
            fields += [
                self.start,
                self.end,
                0,
                self.block_count,
                self._list_to_str(self.block_sizes, sep=','),
                self._list_to_str(self.block_starts, sep=',')
            ]

        return fields

    def to_string(self, all_fields=False):
        data = self.get_data(all_fields=all_fields)
        bed_txt = self._list_to_str(data)
        return bed_txt

    @staticmethod
    def _list_to_str(list_, sep='\t', end=''):
        with io.StringIO() as tmp:
            print(*list_, sep=sep, end=end, file=tmp)
            return tmp.getvalue()

    @classmethod
    def get_union_regions(cls, regions):
        assert len(set(map(itemgetter(0), regions))) == 1, \
            "Not all regions in the same chromosome!"
        assert len(set(map(itemgetter(3), regions))) == 1, \
            "Not all regions at the same strand!"

        chr_ = regions[0][0]
        strand = regions[0][3]

        regions = sorted(
            map(
                lambda r: cls._region(r[1], r[2]),
                regions
            ),
            key=lambda r: r.start
        )

        union_regions = []
        r1 = regions[0]
        for r2 in regions[1:]:
            if r1.end < r2.start:
                union_regions.append(r1)
                r1 = r2
            else:
                if r1.end < r2.end:
                    r1 = cls._region(r1.start, r2.end)
        else:
            union_regions.append(r1)

        union_regions = tuple((chr_, start, end, strand)
                              for start, end in union_regions)
        return union_regions


class BedUtils:
    @classmethod
    def to_regions_df(cls, exons_df):
        if exons_df.empty:
            regions_df = pd.DataFrame([], columns=['regions_id', 'regions'])
        else:
            regions_df = exons_df[['exons_id', 'exons']].assign(
                regions_id=lambda df: df.exons_id,
                regions=lambda df: df.exons.apply(cls._exons_to_regions)
            )[['regions_id', 'regions']]

        return regions_df

    @staticmethod
    def _exons_to_regions(exons):
        regions = list(map(lambda exon: exon.region, exons))
        return regions

    @classmethod
    def to_bed_df(cls, regions_df, union=False):
        if regions_df.empty:
            bed_df = pd.DataFrame([], columns=Bed.BED_TITLE)
        else:
            bed_df = regions_df.apply(cls._to_bed, union=union, axis=1)
            bed_df.columns = Bed.BED_TITLE

        return bed_df

    @classmethod
    def _to_bed(cls, s, union=False):
        bed = Bed(s.regions, name=s.regions_id, union=union)
        return pd.Series(bed.get_data(all_fields=True))


class IntersectBED:
    _dummy_tmpfile = namedtuple('DummyTmpFile', ['name'])

    def __init__(self, bedtools_bin='bedtools', options=''):
        self._bedtools_bin = bedtools_bin

        if isinstance(options, str):
            self._options = options.split()
        else:
            self._options = options

    @classmethod
    def _save_temp(cls, bed):
        if isinstance(bed, str):
            tmp_file = cls._dummy_tmpfile(bed)
        else:
            tmp_file = tp.NamedTemporaryFile(dir='.', prefix='intersectBED.')
            bed.to_csv(tmp_file.name, sep='\t', index=False)

        return tmp_file

    def _generate_cmd(self, bed_file_1, bed_file_2):
        cmd = [self._bedtools_bin, 'intersect'] + self._options
        cmd += ['-a', bed_file_1]
        cmd += ['-b', bed_file_2]

        return cmd

    def intersect(self, bed_1, bed_2):
        bed_tmp_file_1 = self._save_temp(bed_1)
        bed_tmp_file_2 = self._save_temp(bed_2)

        cmd = self._generate_cmd(bed_tmp_file_1.name, bed_tmp_file_2.name)

        with sp.Popen(cmd, stdout=sp.PIPE, encoding='utf-8') as p:
            for line in p.stdout:
                yield line.rstrip('\n').split('\t')



import pandas as pd
from itertools import cycle
from bisect import bisect
from circmimi.bed import IntersectBED, Bed


class RBPBindingSites:
    _titles = Bed.BED_TITLE \
        + tuple(f'{title}_rbp' for title in Bed.BED_TITLE[:6]) \
        + ('overlap',)

    def __init__(self, rbp_file, bedtools_bin="bedtools"):
        self.rbp_file = rbp_file
        self.bedtools_bin = bedtools_bin
        self.intersect_bed = IntersectBED(
            bedtools_bin=self.bedtools_bin,
            options='-wo'
        )

    @staticmethod
    def _recover_blocks(bed_line):
        bed_start = bed_line['start'] + 1
        blockStarts = list(map(int, bed_line['blockStarts'].split(',')))
        blockSizes = list(map(int, bed_line['blockSizes'].split(',')))

        blocks = [Bed._region(bed_start + start, bed_start + start + size - 1)
                  for start, size in zip(blockStarts, blockSizes)]

        return blocks

    @staticmethod
    def _shrink_site_to_region(r, s, offset=0):
        bisect_result = bisect(r, s)

        if bisect_result == 0:
            s_in_r = r.start + offset - 1
        elif bisect_result == 1:
            s_in_r = s
        elif bisect_result == 2:
            s_in_r = r.end + offset

        return s_in_r

    @classmethod
    def _calculate_two_region_overlap(cls, r1, r2):
        start_in_r1 = cls._shrink_site_to_region(r1, r2.start, offset=1)
        end_in_r1 = cls._shrink_site_to_region(r1, r2.end)
        overlap = end_in_r1 - start_in_r1 + 1
        return overlap

    @classmethod
    def _calculate_overlap(cls, bed_line):
        blocks = cls._recover_blocks(bed_line)
        rbp_block = Bed._region(bed_line['start_rbp'] + 1, bed_line['end_rbp'])

        overlap = 0
        for b in blocks:
            overlap += cls._calculate_two_region_overlap(b, rbp_block)

        return overlap

    @classmethod
    def _get_real_overlap(cls, bed_line):
        blockCount = bed_line['blockCount']

        if blockCount == 1:
            overlap = bed_line['overlap']
        elif blockCount > 1:
            overlap = cls._calculate_overlap(bed_line)

        return overlap

    @staticmethod
    def _get_split_rbp_name(idx):
        return lambda df: df['name_rbp'].apply(lambda name: name.split('_')[idx])

    def overlap(self, bed_df):
        intersect_result_df = pd.DataFrame(
            self.intersect_bed.intersect(bed_df, self.rbp_file),
            columns=self._titles,
            dtype='object'
        ).astype(
            {
                'start': int,
                'end': int,
                'blockCount': int,
                'start_rbp': int,
                'end_rbp': int,
                'overlap': int
            }
        ).astype(
            'object'
        ).assign(
            sample_id=self._get_split_rbp_name(0),
            RBP=self._get_split_rbp_name(1)
        ).assign(
            real_overlap=lambda df: df.apply(self._get_real_overlap, axis=1)
        ).assign(
            rbp_region_len=lambda df: df['end_rbp'] - df['start_rbp'],
            total_blocks_len=lambda df: df['blockSizes'].apply(
                lambda sizes: sum(map(int, sizes.split(',')))
            )
        )

        return intersect_result_df

    @staticmethod
    def _get_joined_overlap(group_df):
        joined_overlap = sum(group_df['real_overlap'])
        group_elts = tuple(group_df.index)
        joined_overlap_df = group_df.assign(
            joined_overlap=joined_overlap,
            group_elements=','.join(map(str, group_elts))
        )

        return joined_overlap_df

    @classmethod
    def append_joined_overlap(cls, df):
        groupby_df = df.groupby(
            [
                'name',
                'sample_id'
            ]
        ).apply(
            cls._get_joined_overlap
        ).reset_index(
            [
                'name',
                'sample_id'
            ],
            drop=True
        ).sort_index()

        return groupby_df


class RBPBindingSitesFilters:
    def AGO_overlap_filter(df):
        coverage_df = df[[
            'name',
            'sample_id',
            'RBP',
            'joined_overlap',
            'total_blocks_len'
        ]].drop_duplicates(
        ).assign(
            coverage=lambda df: (df['joined_overlap'] / df['total_blocks_len']).apply(lambda n: round(n, 2))
        )[
            lambda df: df['coverage'] >= 0.8
        ]

        return coverage_df

    def RBP_overlap_filter(df):
        coverage_df = df[[
            'name',
            'chr_rbp',
            'start_rbp',
            'end_rbp',
            'strand_rbp',
            'sample_id',
            'RBP',
            'real_overlap',
            'rbp_region_len'
        ]].assign(
            coverage=lambda df: (df['real_overlap']/df['rbp_region_len']).apply(lambda n: round(n, 2))
        )[
            lambda df: df['coverage'] >= 0.8
        ]

        return coverage_df


class PosMap:
    def __init__(self, genomic_regions):
        self._regions = genomic_regions

        self._init_lens(genomic_regions)

    def _init_lens(self, regions):
        self._lens = list(map(lambda r: r[2] - r[1] + 1, regions))

    @staticmethod
    def get_real_pos(rel_pos, genomic_region):
        chr_, start, end, strand = genomic_region

        if strand == '+':
            pos = start + rel_pos - 1
        elif strand == '-':
            pos = end - rel_pos + 1

        return (chr_, pos, strand)

    def regions_iter(self):
        return cycle(zip(self._lens, self._regions))

    @staticmethod
    def _to_region(pos1, pos2):
        c1, p1, s1 = pos1
        c2, p2, s2 = pos2

        assert c1 == c2
        assert s1 == s2

        p1, p2 = sorted([p1, p2])

        return (c1, p1, p2, s1)

    def get_real_blocks(self, rel_start, rel_end):
        real_blocks = []
        regions = self.regions_iter()

        l, r = next(regions)

        while rel_start > l:
            rel_start -= l
            rel_end -= l
            l, r = next(regions)

        start_pos = self.get_real_pos(rel_start, r)

        if rel_end <= l:
            end_pos = self.get_real_pos(rel_end, r)
            real_blocks.append(self._to_region(start_pos, end_pos))
        else:
            while rel_end > l:
                end_pos = self.get_real_pos(l - 1, r)
                real_blocks.append(self._to_region(start_pos, end_pos))

                rel_end -= l
                l, r = next(regions)
                start_pos = self.get_real_pos(0, r)

            end_pos = self.get_real_pos(rel_end, r)
            real_blocks.append(self._to_region(start_pos, end_pos))

        return tuple(real_blocks)



#! /usr/bin/env python

"""
A python script to get the sequences consist of the flanking +-100bp
 sequences on circRNA from the circRNA junction.

Note:
  The `bedtools` is required.
  We use `bedtools getfasta` to get the needed sequences from the genome.

Usage:
  get_flanking_seq.py [Genome] [CircRNAs] [Output]

  Genome:
    The reference genome file.

  CircRNAs:
    CircRNAs data in the 4-columns format with (chr, pos1, pos2, strand).

  Output:
    The output file in FASTA format.
"""

import sys
import io
import re
import tempfile as tp
import subprocess as sp
import logging


logging.basicConfig(
    format="{asctime} - {message}",
    level=logging.INFO,
    style='{'
)


class Bed:
    BED_TITLE = (
        'chr',
        'start',
        'end',
        'name',
        'score',
        'strand',
        'thickStart',
        'thickEnd',
        'itemRGB',
        'blockCount',
        'blockSizes',
        'blockStarts'
    )

    def __init__(self, regions, name='.'):
        self.regions = regions
        self.name = name

        self._parse_regions(regions)

    def _parse_regions(self, regions):
        self._parse_region(regions[0])
        for region in regions[1:]:
            self._add_block(region)

        if self.strand == '-':
            self._reverse_region()

    def _parse_region(self, region):
        self.chrom = region[0]
        self.start = int(region[1]) - 1
        self.end = int(region[2])
        self.strand = region[3]

        self.block_count = 1
        self.block_sizes = [self.end - self.start]
        self.block_starts = [0]

    def _add_block(self, region):
        other = Bed([region])

        assert self.strand == other.strand

        self.block_count += other.block_count
        self.block_sizes += other.block_sizes

        all_starts = [
            start + self.start for start in self.block_starts
        ] + [other.start]

        if other.start < self.start:
            self.start = other.start

        self.block_starts = [start - self.start for start in all_starts]

        if other.end > self.end:
            self.end = other.end

    def _reverse_region(self):
        self.block_sizes = list(reversed(self.block_sizes))
        self.block_starts = list(reversed(self.block_starts))

    def get_data(self, all_fields=False):
        fields = [
            self.chrom,
            self.start,
            self.end,
            self.name,
            '.',
            self.strand
        ]

        if all_fields:
            fields += [
                self.start,
                self.end,
                0,
                self.block_count,
                self._list_to_str(self.block_sizes, sep=','),
                self._list_to_str(self.block_starts, sep=',')
            ]

        return fields

    def to_string(self, all_fields=False):
        data = self.get_data(all_fields=all_fields)
        bed_txt = self._list_to_str(data)
        return bed_txt

    @staticmethod
    def _list_to_str(list_, sep='\t', end=''):
        with io.StringIO() as tmp:
            print(*list_, sep=sep, end=end, file=tmp)
            return tmp.getvalue()


def get_fasta(bed_file, ref_file, use_blocks=False, bedtools_bin='bedtools'):
    cmd = [bedtools_bin, 'getfasta']
    cmd += ['-fi', ref_file]
    cmd += ['-bed', bed_file]
    cmd += ['-fo', '/dev/stdout']
    cmd += ['-name', '-s']

    if use_blocks:
        cmd += ['-split']

    res = sp.run(cmd, stdout=sp.PIPE, encoding='utf-8')

    return res.stdout


class CircRNAs:
    def __init__(self, chr_, pos1, pos2, strand, circ_id=None):
        self.chr_ = chr_
        self.pos1 = int(pos1)
        self.pos2 = int(pos2)
        self.strand = strand
        self.circ_id = circ_id

        self.get_donor_acceptor()

    def get_donor_acceptor(self):
        pos1, pos2 = sorted([self.pos1, self.pos2])

        if self.strand == '+':
            self.donor_site = pos2
            self.acceptor_site = pos1
        elif self.strand == '-':
            self.donor_site = pos1
            self.acceptor_site = pos2

    def get_flanking_region(self):
        donor_flanking = self._get_flanking_region_of_donor()
        acceptor_flanking = self._get_flanking_region_of_acceptor()
        return [donor_flanking, acceptor_flanking]

    def _get_flanking_region_of_donor(self):
        if self.strand == '+':
            donor_start = self.donor_site - 99
            donor_end = self.donor_site

        elif self.strand == "-":
            donor_start = self.donor_site
            donor_end = self.donor_site + 99

        donor_flanking = [
            self.chr_,
            donor_start,
            donor_end,
            self.strand
        ]

        return donor_flanking

    def _get_flanking_region_of_acceptor(self):
        if self.strand == '+':
            acceptor_start = self.acceptor_site
            acceptor_end = self.acceptor_site + 99

        elif self.strand == "-":
            acceptor_start = self.acceptor_site - 99
            acceptor_end = self.acceptor_site

        acceptor_flanking = [
            self.chr_,
            acceptor_start,
            acceptor_end,
            self.strand
        ]

        return acceptor_flanking

    @property
    def id(self):
        if self.circ_id:
            return self.circ_id
        else:
            return f"{self.chr_}:{self.donor_site}|{self.acceptor_site}({self.strand})"


def get_flanking_seq(genome_file, circRNAs_file, out_file):
    logging.info("Start to get the flanking sequences.")

    tmp_bed = tp.NamedTemporaryFile(
        prefix='flanking_region.',
        suffix='.bed',
        dir='.'
    )

    with open(circRNAs_file) as circRNA_data, \
            open(tmp_bed.name, 'w') as bed_out:
        for line in circRNA_data:
            circRNA = CircRNAs(*line.rstrip('\n').split('\t'))
            flanking_region = circRNA.get_flanking_region()
            region_bed = Bed(flanking_region, name=circRNA.id)
            print(region_bed.to_string(all_fields=True), file=bed_out)

    fasta_data = get_fasta(tmp_bed.name, genome_file, use_blocks=True)

    with open(out_file, 'w') as out:
        out.write(
            re.sub(r'(?<=\([+-]\))\([+-]\)$', '', fasta_data, flags=re.M)
        )

    tmp_bed.close()

    logging.info("Process completed!")


def print_usage():
    print(__doc__, file=sys.stderr)


def cli():
    if len(sys.argv) < 4:
        print_usage()
        exit(1)

    get_flanking_seq(*sys.argv[1:])


if __name__ == "__main__":
    cli()



#! /usr/bin/env python

import argparse
from collections import namedtuple


TITLES = [
    'circRNA_id',
    'region1',
    'region2',
    'matches',
    'alignment_len',
    'mismatch',
    'gapopen',
    'alignment_start_1',
    'alignment_end_1',
    'alignment_start_2',
    'alignment_end_2',
    'E_value',
    'bit_score',
    'region1_type',
    'region2_type'
]


class RCSReader:
    RCS_data = namedtuple('RCS_data', TITLES)

    def __init__(self, RCS_result_file):
        self._RCS_result_file = RCS_result_file
        self._reader = self._RCS_data_reader()

    def _RCS_data_reader(self):
        self._title = self._RCS_result_file.readline()

        for line in self._RCS_result_file:
            data = line.rstrip('\n').split('\t')
            data = self.RCS_data(*data)
            yield data

    def __iter__(self):
        return self._reader


def RCS_filter(RCS_result,
               min_matches=80,
               min_alignment_len=50,
               min_bitscore=100):

    for data in RCS_result:
        if data.matches != '':
            if (float(data.matches) > min_matches) \
                and (int(data.alignment_len) > min_alignment_len) \
                    and (float(data.bit_score) > min_bitscore):

                yield data


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('-m', '--min-matches', type=float, default=80)
    parser.add_argument('-l', '--min-aln-len', type=int, default=50)
    parser.add_argument('-b', '--min-bitscore', type=float, default=100)
    parser.add_argument('RCS_file', type=argparse.FileType('r'))

    return parser


def cli():
    parser = create_parser()
    args = parser.parse_args()

    print(*TITLES, sep='\t')

    reader = RCSReader(args.RCS_file)
    reader = RCS_filter(
        reader,
        min_matches=args.min_matches,
        min_alignment_len=args.min_aln_len,
        min_bitscore=args.min_bitscore
    )

    for data in reader:
        print(*data, sep='\t')


if __name__ == "__main__":
    cli()



#! /usr/bin/env python

import argparse
from itertools import groupby
from collections import namedtuple, Counter


INPUT_TITLES = [
    'circRNA_id',
    'region1',
    'region2',
    'matches',
    'alignment_len',
    'mismatch',
    'gapopen',
    'alignment_start_1',
    'alignment_end_1',
    'alignment_start_2',
    'alignment_end_2',
    'E_value',
    'bit_score',
    'region1_type',
    'region2_type'
]


OUTPUT_TITLES = [
    'circRNA_id',
    '#RCS across flanking sequences',
    '#RCS within the flanking sequence (the donor side)',
    '#RCS within the flanking sequence (the acceptor side)',
    '#RCS_across-#RCS_within>=1 (yes: 1; no: 0)'
]


class RCSReader:
    RCS_data = namedtuple('RCS_data', INPUT_TITLES)

    def __init__(self, RCS_result_file):
        self._RCS_result_file = RCS_result_file
        self._reader = self._RCS_data_reader()

    def _RCS_data_reader(self):
        self._title = self._RCS_result_file.readline()

        for line in self._RCS_result_file:
            data = line.rstrip('\n').split('\t')
            data = self.RCS_data(*data)
            yield data

    def __iter__(self):
        return self._reader


def circRNA_id_reader(circ_file):
    if circ_file is not None:
        for line in circ_file:
            data = line.rstrip('\n').split('\t')
            chr_ = data[0]
            pos1, pos2 = sorted(map(int, data[1:3]))
            strand = data[3]

            if len(data) >= 5:
                circ_id = data[4]
            else:
                if strand == '+':
                    circ_id = f"{chr_}:{pos2}|{pos1}({strand})"
                elif strand == '-':
                    circ_id = f"{chr_}:{pos1}|{pos2}({strand})"

            yield circ_id


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--circ_file', type=argparse.FileType('r'),
                        help="4-columns TSV file: (chr, pos1, pos2, strand)")
    parser.add_argument('RCS_file', type=argparse.FileType('r'))

    return parser


def cli():
    parser = create_parser()
    args = parser.parse_args()

    print(*OUTPUT_TITLES, sep='\t')

    reader = RCSReader(args.RCS_file)
    group_reader = groupby(reader, key=lambda data: data.circRNA_id)

    circRNA_ids_reader = circRNA_id_reader(args.circ_file)

    for circRNA_id, RCS_gp in group_reader:

        for circ_id in circRNA_ids_reader:
            if circ_id != circRNA_id:
                print(circ_id, 0, 0, 0, sep='\t')
            else:
                break

        RCS_gp = list(RCS_gp)
        RCS_types = Counter(
            [(data.region1_type, data.region2_type) for data in RCS_gp]
        )

        across = RCS_types[('donor', 'acceptor')]
        within_donor = RCS_types[('donor', 'donor')]
        within_acceptor = RCS_types[('acceptor', 'acceptor')]
        across_minus_within = across - (within_donor + within_acceptor)
        across_minus_within_le_one = int(across_minus_within >= 1)

        print(
            circRNA_id,
            across,
            within_donor,
            within_acceptor,
            across_minus_within_le_one,
            sep='\t'
        )
    else:
        for circ_id in circRNA_ids_reader:
            print(circ_id, 0, 0, 0, 0, sep='\t')


if __name__ == "__main__":
    cli()



#! /usr/bin/env python

"""
A python script to check for reads if there exist colinear alignments
 or multiple alignments.

External tools:
  - blat
  - mp_blat.py

Usage:
  checkAA_reads.py -rg [Genome] -ro [OtherRefs] -p [NumProc] [Reads]
"""


import argparse
import tempfile as tp
import subprocess as sp
import io
import csv
import textwrap
import re
from itertools import groupby, chain
from functools import partial
from collections import namedtuple, deque


class Blat:
    def __init__(self, work_dir='.', blat_bin='blat', blat_opts="",
                 num_proc=1, mp_blat_bin='mp_blat.py'):
        self.work_dir = work_dir
        self.blat_bin = blat_bin
        self.blat_opts = blat_opts
        self.num_proc = num_proc
        self.mp_blat_bin = mp_blat_bin

    def _run(self, ref_file, fa_file):
        with tp.NamedTemporaryFile(dir=self.work_dir) as tmp_out:
            cmd = self._generate_cmd(ref_file, fa_file, tmp_out.name)
            sp.run(cmd)

            result = BlatPsl(tmp_out.read().decode('utf-8'))
            return result

    def _generate_cmd(self, ref_file, fa_file, out_file):
        cmd = [self.mp_blat_bin, ref_file, fa_file, out_file]
        cmd += ['-p', str(self.num_proc)]
        cmd += ['--tmp_path', self.work_dir]
        cmd += ['--blat_bin', self.blat_bin]
        cmd += ['--blat_opt', self.blat_opts]
        return cmd

    def __call__(self, ref_file, fa_file):
        return self._run(ref_file, fa_file)


class BlatPsl:
    PSL_TITLE = (
        'matches',
        'misMatches',
        'repMatches',
        'nCount',
        'qNumInsert',
        'qBaseInsert',
        'tNumInsert',
        'tBaseInsert',
        'strand',
        'qName',
        'qSize',
        'qStart',
        'qEnd',
        'tName',
        'tSize',
        'tStart',
        'tEnd',
        'blockCount',
        'blockSizes',
        'qStarts',
        'tStarts'
    )
    _str_cols = ['strand', 'qName', 'tName']
    _list_cols = ['blockSizes', 'qStarts', 'tStarts']
    _psl = namedtuple('Psl', PSL_TITLE)

    def __init__(self, psl_text):
        self.psl_text = psl_text
        self.psl_data = self._parse(psl_text)

    @classmethod
    def _parse(cls, psl_text):
        no_header_psl = cls._remove_header(psl_text)

        with io.StringIO() as tmp_out:
            print(*cls.PSL_TITLE, sep='\t', file=tmp_out)
            tmp_out.write(no_header_psl)
            tmp_out.seek(0)

            psl_data = [cls._psl(**cls._correct_column_type(data))
                        for data in csv.DictReader(tmp_out, delimiter='\t')]

            return psl_data

    @staticmethod
    def _remove_header(psl_text):
        if psl_text.startswith('psLayout'):
            no_header_psl = psl_text.split('\n', 5)[5]
        else:
            no_header_psl = psl_text

        return no_header_psl

    @classmethod
    def _correct_column_type(cls, data):
        corrected_data = {}
        for k, v in data.items():
            if k in cls._str_cols:
                new_v = v
            elif k in cls._list_cols:
                new_v = tuple(map(int, v.rstrip(',').split(',')))
            else:
                new_v = int(v)

            corrected_data[k] = new_v

        return corrected_data


class PslFilters:
    @staticmethod
    def _is_colinear(match, threshold=0.8):
        colinear_score = (match.matches + match.repMatches) / match.qSize
        return colinear_score > threshold

    @classmethod
    def colinear_filter(cls, psl_data, threshold=0.8):
        for match in psl_data:
            if cls._is_colinear(match, threshold=threshold):
                yield match

    @staticmethod
    def _is_chimera(psl_data, threshold):
        chimera = PslChimera()
        chimera.parse(psl_data)

        if not chimera.linear:
            if (threshold.check(chimera.five_end) is True) and \
               (threshold.check(chimera.three_end) is True):
                return True

        return False

    @classmethod
    def chimera_filter(cls, psl_data, min_matches=30, min_diff=3):
        threshold = _Threshold(min_matches=min_matches, min_diff=min_diff)
        _is_chimera = partial(cls._is_chimera, threshold=threshold)

        for _, gp in groupby(psl_data, key=lambda match: match.qName):
            gp = list(gp)
            if _is_chimera(gp):
                yield gp


class _Threshold:
    def __init__(self, min_matches, min_diff):
        self.min_matches = min_matches
        self.min_diff = min_diff

    def check(self, matches):
        num_matches = len(matches)
        if num_matches == 0:
            best_match_score = 0
            second_best_score = 0
        elif num_matches == 1:
            best_match_score = matches[0].matches
            second_best_score = 0
        elif num_matches == 2:
            best_match_score = matches[0].matches
            second_best_score = matches[1].matches

        if best_match_score < self.min_matches:
            return False

        if best_match_score - second_best_score < self.min_diff:
            return False

        return True


class PslChimera:
    def __init__(self):
        self._linear = False
        self._five_end = deque(maxlen=2)
        self._three_end = deque(maxlen=2)

    def parse(self, psl_data):
        for match in psl_data:
            if self._is_colinear(match):
                self._linear = True

            if self._is_five_end(match):
                self._add_best_match(self._five_end, match)

            if self._is_three_end(match):
                self._add_best_match(self._three_end, match)

    @property
    def linear(self):
        return self._linear

    @property
    def five_end(self):
        return self._five_end

    @property
    def three_end(self):
        return self._three_end

    @staticmethod
    def _is_colinear(match):
        return (match.qStart < 10) and (match.qSize - match.qEnd < 10)

    @staticmethod
    def _is_five_end(match):
        return match.qStart < 10

    @staticmethod
    def _is_three_end(match):
        return match.qSize - match.qEnd < 10

    @staticmethod
    def _is_better_match(match, other):
        return match.matches >= other.matches

    @classmethod
    def _add_best_match(cls, q, match):
        len_q = len(q)

        if len_q == 0:
            q.append(match)
        elif len_q == 1:
            if cls._is_better_match(match, q[0]):
                q.appendleft(match)
            else:
                q.append(match)
        else:
            if cls._is_better_match(match, q[0]):
                q.appendleft(match)
            elif cls._is_better_match(match, q[1]):
                q.appendleft(match)
                q.reverse()
            else:
                pass


class PslUtils:
    @staticmethod
    def get_uniq_qname(psl_data):
        uniq_qnames = sorted(set(map(lambda match: match.qName, psl_data)))
        return uniq_qnames


class AmbAlnChecker:
    def __init__(self,
                 ref_genome,
                 ref_others,
                 work_dir='.',
                 num_proc=1,
                 blat_bin='blat',
                 mp_blat_bin='mp_blat.py'):

        self.ref_genome = ref_genome
        self.ref_others = ref_others
        self.work_dir = work_dir
        self.num_proc = num_proc
        self.blat_bin = blat_bin
        self.mp_blat_bin = mp_blat_bin

        self.blat = Blat(
            work_dir=self.work_dir,
            num_proc=self.num_proc,
            blat_bin=self.blat_bin,
            mp_blat_bin=self.mp_blat_bin
        )
        self.web_blat = Blat(
            work_dir=self.work_dir,
            num_proc=self.num_proc,
            blat_opts="-tileSize=9 -stepSize=9 -repMatch=32768",
            blat_bin=self.blat_bin,
            mp_blat_bin=self.mp_blat_bin
        )

    @property
    def infos(self):
        infos = {
            'ref_genome': self.ref_genome,
            'ref_others': self.ref_others,
            'work_dir': self.work_dir,
            'num_proc': self.num_proc,
            'blat_bin': self.blat_bin,
            'mp_blat_bin': self.mp_blat_bin
        }
        return infos

    def check(self, reads_file):
        result = AmbAlnResult(reads_file=reads_file)
        result.set_infos(**self.infos)

        # run blat
        result._psl_rG_1 = self.blat(self.ref_genome, reads_file)
        result._psl_rG_2 = self.web_blat(self.ref_genome, reads_file)
        result._psl_rO_1 = self.blat(self.ref_others, reads_file)
        result._psl_rO_2 = self.web_blat(self.ref_others, reads_file)

        # evaluate
        # colinear part
        result._psl_rG_1_colinear = PslFilters.colinear_filter(result._psl_rG_1.psl_data)
        result._psl_rG_2_colinear = PslFilters.colinear_filter(result._psl_rG_2.psl_data)
        result._psl_rO_1_colinear = PslFilters.colinear_filter(result._psl_rO_1.psl_data)
        result._psl_rO_2_colinear = PslFilters.colinear_filter(result._psl_rO_2.psl_data)

        result._psl_rG_1_colinear_ids = PslUtils.get_uniq_qname(result._psl_rG_1_colinear)
        result._psl_rG_2_colinear_ids = PslUtils.get_uniq_qname(result._psl_rG_2_colinear)
        result._psl_rO_1_colinear_ids = PslUtils.get_uniq_qname(result._psl_rO_1_colinear)
        result._psl_rO_2_colinear_ids = PslUtils.get_uniq_qname(result._psl_rO_2_colinear)

        result.colinear_ids = sorted(set(chain(
            result._psl_rG_1_colinear_ids,
            result._psl_rG_2_colinear_ids,
            result._psl_rO_1_colinear_ids,
            result._psl_rO_2_colinear_ids
        )))

        # multiple hits part
        result._psl_rG_1_chimera = PslFilters.chimera_filter(result._psl_rG_1.psl_data)
        result._psl_rG_2_chimera = PslFilters.chimera_filter(result._psl_rG_2.psl_data)
        result._psl_rG_1_chimera_ids = PslUtils.get_uniq_qname(chain.from_iterable(result._psl_rG_1_chimera))
        result._psl_rG_2_chimera_ids = PslUtils.get_uniq_qname(chain.from_iterable(result._psl_rG_2_chimera))

        result._psl_rG_1_multiple_hits = list(filter(
            lambda m: m.qName not in result._psl_rG_1_colinear_ids + result._psl_rG_1_chimera_ids,
            result._psl_rG_1.psl_data
        ))
        result._psl_rG_2_multiple_hits = list(filter(
            lambda m: m.qName not in result._psl_rG_2_colinear_ids + result._psl_rG_2_chimera_ids,
            result._psl_rG_2.psl_data
        ))
        result._psl_rG_1_multiple_hits_ids = PslUtils.get_uniq_qname(result._psl_rG_1_multiple_hits)
        result._psl_rG_2_multiple_hits_ids = PslUtils.get_uniq_qname(result._psl_rG_2_multiple_hits)

        result.multiple_hits_ids = sorted(
            set(chain(
                result._psl_rG_1_multiple_hits_ids,
                result._psl_rG_2_multiple_hits_ids
            )).difference(
                set(result.colinear_ids)
            )
        )

        result.update_result()

        return result


class AmbAlnResult:
    _id_in_fasta = re.compile(r'^>([^ ]+).*\n$')

    def __init__(self, reads_file):
        self.reads_file = reads_file
        self.infos = {}

        self._init_result()

    def _init_result(self):
        self.read_ids = self._get_read_ids(self.reads_file)
        self.result = {read_id: [0, 0, 0] for read_id in self.read_ids}

    @classmethod
    def _get_read_ids(cls, reads_file):
        read_ids = []
        with open(reads_file) as f_in:
            for line in f_in:
                if line.startswith('>'):
                    read_ids.append(re.match(cls._id_in_fasta, line).group(1))
        return read_ids

    def set_infos(self, **kwargs):
        self.infos.update(kwargs)

    def update_result(self):
        for id_ in self.colinear_ids:
            self.result[id_][0] = 1

        for id_ in self.multiple_hits_ids:
            self.result[id_][1] = 1

        for id_, checking_result in self.result.items():
            if (checking_result[0] == 1) or (checking_result[1] == 1):
                self.result[id_][2] = 1

    def save_result(self, out_file):
        with open(out_file, 'w') as out:
            csv_writer = csv.writer(out, delimiter='\t')

            csv_writer.writerow(
                [
                    'circRNA_id',
                    'with an alternative co-linear explanation',
                    'with multiple_hits',
                    'alignment ambiguity (with an alternative co-linear explanation or multiple hits)'
                ]
            )

            for read_id, check_items in self.result.items():
                csv_writer.writerow([read_id] + check_items)


def create_parser():

    parser = argparse.ArgumentParser(

        description=textwrap.dedent("""
            A python script to check for reads if there exist colinear alignments
             or multiple alignments.
            """),
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument('-rg', dest='ref_genome', required=True)
    parser.add_argument('-ro', dest='ref_others', required=True)
    parser.add_argument('in_file', help='input reads file.')
    parser.add_argument('out_file')
    parser.add_argument('-p', dest='num_proc', type=int, default=1)

    return parser


def cli():
    parser = create_parser()
    args = parser.parse_args()

    checker = AmbAlnChecker(
        ref_genome=args.ref_genome,
        ref_others=args.ref_others,
        num_proc=args.num_proc
    )

    result = checker.check(args.in_file)
    result.save_result(args.out_file)


if __name__ == "__main__":
    cli()



import logging
import pandas as pd
from circmimi.circ import CircEvents
from circmimi.bed import BedUtils
from circmimi.seq import Seq
from circmimi.miranda import get_binding_sites, MirandaUtils
from circmimi.rbp import PosMap, RBPBindingSites, RBPBindingSitesFilters
from circmimi.stats import do_the_hypergeometric_test


logger = logging.getLogger(__name__)


class Circmimi:
    def __init__(self,
                 anno_db_file,
                 ref_file,
                 mir_ref_file,
                 mir_target_file,
                 AGO_binding_file=None,
                 RBP_binding_file=None,
                 RBP_target_file=None,
                 other_ref_file=None,
                 work_dir='.',
                 num_proc=1,
                 pv_filter=True,
                 miranda_options=None):

        self.anno_db_file = anno_db_file
        self.ref_file = ref_file
        self.mir_ref_file = mir_ref_file
        self.mir_target_file = mir_target_file
        self.AGO_binding_file = AGO_binding_file
        self.RBP_binding_file = RBP_binding_file
        self.RBP_target_file = RBP_target_file
        self.other_ref_file = other_ref_file
        self.work_dir = work_dir
        self.num_proc = num_proc
        self.pv_filter = pv_filter
        self.miranda_options = miranda_options

        self.circ_events = None
        self.uniq_exons_df = None
        self.bed_df = None
        self.seq_df = None
        self.miranda_df = None
        self.grouped_res_df = None

        self.mir_target_db = get_mir_target_db(self.mir_target_file)

        if self.AGO_binding_file:
            self.AGO_binding_sites = RBPBindingSites(self.AGO_binding_file)
            self.check_AGO_support = True
        else:
            self.check_AGO_support = False

        # if self.RBP_binding_file:
        #     self.RBP_binding_sites = RBPBindingSites(self.RBP_binding_file)
        #     self.do_circRNA_RBP = True
        # else:
        #     self.do_circRNA_RBP = False

        # if self.RBP_target_file:
        #     self.RBP_target_db = pd.read_csv(
        #         self.RBP_target_file,
        #         sep='\t',
        #         dtype='object'
        #     )
        #     self.do_RBP_mRNA = True
        # else:
        #     self.do_RBP_mRNA = False
        self.do_circRNA_RBP = False
        self.do_RBP_mRNA = False

    def _init_results(self):
        res_df_columns = ['ev_id'] + list(CircEvents.INPUT_COLUMNS)
        self.res_df = pd.DataFrame([], columns=res_df_columns)
        self.circ_target_df_with_pv = pd.DataFrame([])


    def run(self, circ_file):
        logger.info('loading circRNAs')
        self.circ_events = CircEvents(circ_file)

        logger.info('checking gene annotation for these circRNAs')
        self.circ_events.check_annotation(self.anno_db_file)

        if self.other_ref_file is not None:
            logger.info('checking ambiguous alignments')
            self.circ_events.check_ambiguous(
                self.anno_db_file,
                self.ref_file,
                self.other_ref_file,
                work_dir=self.work_dir,
                num_proc=self.num_proc
            )

        logger.info('getting all possible isoforms of these circRNAs')

        if self.circ_events.clear_df.empty:
            logger.info("No isoforms were found!")
            self._init_results()
            return

        self.uniq_exons_df = self.circ_events.clear_anno_df.pipe(
            self._get_uniq_exons
        )
        self.uniq_exons_regions_df = self.uniq_exons_df.pipe(
            BedUtils.to_regions_df
        )
        self.bed_df = self.uniq_exons_regions_df.pipe(
            BedUtils.to_bed_df
        )

        self.seq_df = self.bed_df.pipe(
            Seq.get_extended_seq,
            ref_file=self.ref_file
        )

        self.pos_map_db = {
            regions_id: PosMap(regions)
            for regions_id, regions in self.uniq_exons_regions_df.values
        }

        # miRNAs part
        logger.info('predicting miRNA-binding sites on circRNAs')
        self.miranda_df = self.seq_df.pipe(
            get_binding_sites,
            mir_ref_file=self.mir_ref_file,
            work_dir=self.work_dir,
            num_proc=self.num_proc,
            miranda_options=self.miranda_options
        ).pipe(
            MirandaUtils.append_exons_len,
            exons_len_df=self.uniq_exons_df[['exons_id', 'total_len']]
        ).pipe(
            MirandaUtils.remove_redundant_result
        ).pipe(
            MirandaUtils.append_cross_boundary
        ).pipe(
            MirandaUtils.append_ev_id,
            exons_ev_id_df=self.uniq_exons_df[['exons_id', 'ev_id']]
        ).pipe(
            MirandaUtils.append_merged_aln
        ).pipe(
            MirandaUtils.generate_uniq_id,
            column_name='aln'
        ).drop(
            'aln',
            axis=1
        ).pipe(
            MirandaUtils.get_genomic_position,
            pos_map_db=self.pos_map_db
        ).pipe(
            MirandaUtils.generate_uniq_id,
            column_name='genomic_regions'
        )

        # AGO overlap
        if self.check_AGO_support:
            logger.info('filtering AGO-supported miRNA-binding sites')
            logger.debug('getting miRNA_binding_sites_bed')
            self.miRNA_binding_sites_bed = self.miranda_df[[
                'genomic_regions_id',
                'genomic_regions'
            ]].drop_duplicates(
            ).reset_index(
                drop=True
            ).rename(
                {
                    'genomic_regions_id': 'regions_id',
                    'genomic_regions': 'regions'
                },
                axis=1
            ).pipe(
                BedUtils.to_bed_df,
                union=True
            )

            logger.debug('getting AGO_overlap_raw_data')
            self.AGO_overlap_raw_data = self.AGO_binding_sites.overlap(
                self.miRNA_binding_sites_bed
            ).pipe(
                RBPBindingSites.append_joined_overlap
            )

            logger.debug('getting AGO_overlap')
            self.AGO_overlap = self.AGO_overlap_raw_data.pipe(
                RBPBindingSitesFilters.AGO_overlap_filter
            )

            logger.debug('getting AGO_overlap_count')
            self.AGO_overlap_count = self.AGO_overlap[[
                'name',
                'sample_id'
            ]].groupby(
                'name'
            ).count(
            ).reset_index(
            ).rename(
                {
                    'name': 'genomic_regions_id',
                    'sample_id': 'AGO_support'
                },
                axis=1
            ).astype(
                {
                    'AGO_support': 'object'
                }
            )

            logger.debug('merge back to miranda_df')
            self.miranda_df = self.miranda_df.merge(
                self.AGO_overlap_count,
                on='genomic_regions_id',
                how='left'
            ).fillna(
                {
                    'AGO_support': 0
                }
            ).assign(
                AGO_support_yn=lambda df: (df['AGO_support'] > 0).apply(int)
            )

        logger.debug('grouping results (miranda_df)')
        self.grouped_res_df = MirandaUtils.get_grouped_results(
            self.miranda_df,
            with_AGO=self.check_AGO_support
        )

        # RBP part
        if self.do_circRNA_RBP:
            logger.info('predicting RBP-binding sites on circRNAs')
            self.union_bed_df = self.uniq_exons_regions_df.pipe(
                BedUtils.to_bed_df,
                union=True
            )

            self.RBP_overlap_raw_data = self.RBP_binding_sites.overlap(
                self.union_bed_df
            )
            self.RBP_overlap = self.RBP_overlap_raw_data.pipe(
                RBPBindingSitesFilters.RBP_overlap_filter
            ).merge(
                self.uniq_exons_df[['exons_id', 'ev_id']],
                left_on='name',
                right_on='exons_id',
                how='left'
            ).drop(
                'exons_id',
                axis=1
            )
            self.RBP_overlap_count = self.RBP_overlap[[
                'ev_id',
                'RBP',
                'chr_rbp',
                'start_rbp',
                'end_rbp',
                'strand_rbp',
                'sample_id'
            ]].drop_duplicates(
            ).groupby(
                [
                    'ev_id',
                    'RBP',
                    'chr_rbp',
                    'start_rbp',
                    'end_rbp',
                    'strand_rbp'
                ]
            ).agg(
                {
                    'sample_id': 'nunique'
                }
            ).reset_index(
            ).assign(
                count=1
            ).groupby(
                [
                    'ev_id',
                    'RBP'
                ]
            ).agg(
                {
                    'sample_id': 'max',
                    'count': 'sum'
                }
            ).reset_index(
            ).rename(
                {
                    'sample_id': 'MaxRbpExpNum',
                    'count': 'num_RBP_binding_sites'
                },
                axis=1
            )

        # final result table
        logger.info('getting final results')
        logger.debug('getting res_df')
        self.res_df = self.circ_events.clear_df.pipe(
            debug_log,
            msg='merging res_df'
        ).merge(
            self.grouped_res_df,
            on='ev_id',
            how='inner'
        ).pipe(
            debug_log,
            msg='merging mir_target'
        ).merge(
            self.mir_target_db,
            on='mirna',
            how='inner'
        ).sort_values(
            [
                'ev_id',
                'mirna',
                'target_gene'
            ]
        ).reset_index(drop=True)

        self.category_df = self.res_df.pipe(self._get_category_df, to_binary=True)
        self.res_df = pd.concat([self.res_df, self.category_df], axis=1)

        if self.do_circRNA_RBP:
            logger.debug('getting RBP_res_df')
            logger.debug('merging gene_symbol')
            self.RBP_res_df = self.circ_events.clear_df.pipe(
                debug_log,
                msg='merging RBP_overlap'
            ).merge(
                self.RBP_overlap_count,
                on='ev_id',
                how='left'
            )

            if self.do_RBP_mRNA:
                logger.debug('merging RBP_target')
                self.RBP_res_df = self.RBP_res_df.merge(
                    self.RBP_target_db,
                    on='RBP',
                    how='inner'
                )
        else:
            self.RBP_res_df = None

        # calculate P-value
        logger.info('calculating the P-values for the interactions of circRNAs and target genes')
        self.circ_mir_target_df = self.res_df[['circ_id', 'mirna', 'target_gene']].drop_duplicates().reset_index(drop=True)
        self.circ_target_df_with_pv = do_the_hypergeometric_test(
            self.circ_mir_target_df,
            self.mir_ref_file,
            self.mir_target_db
        )
        self.res_df = self.res_df.merge(
            self.circ_target_df_with_pv[[
                'circ_id',
                'target_gene',
                'p_value',
                'bh_corrected_p_value',
                'bonferroni_corrected_p_values',
            ]],
            on=['circ_id', 'target_gene'],
            how='left'
        )

        # only retain interactions with 'bh_corrected_p_value < 0.05'
        if self.pv_filter:
            self.res_df = self.res_df[self.res_df['bh_corrected_p_value'] < 0.05]

        # submit summary
        logger.info('generating summary')
        circ_miRNA_count = self.res_df[['ev_id', 'mirna']].drop_duplicates().rename(
            {
                'mirna': '#circRNA_miRNA'
            },
            axis=1
        ).groupby(
            'ev_id'
        ).agg(
            'count'
        ).pipe(
            self.circ_events.expand_to_all_events,
            fillna_value=0
        ).astype('int')
        self.circ_events.submit_to_summary(circ_miRNA_count, type_='summary')

        circ_mRNA_count = self.res_df[['ev_id', 'target_gene']].drop_duplicates().rename(
            {
                'target_gene': '#circRNA_mRNA'
            },
            axis=1
        ).groupby(
            'ev_id'
        ).agg(
            'count'
        ).pipe(
            self.circ_events.expand_to_all_events,
            fillna_value=0
        ).astype('int')
        self.circ_events.submit_to_summary(circ_mRNA_count, type_='summary')

        total_count = self.res_df[['ev_id', 'mirna', 'target_gene']].drop_duplicates().assign(
            circRNA_miRNA_mRNA=1
        ).drop(
            ['mirna', 'target_gene'],
            axis=1
        ).groupby(
            'ev_id'
        ).agg(
            'count'
        ).pipe(
            self.circ_events.expand_to_all_events,
            fillna_value=0
        ).astype('int').rename(
            {
                'circRNA_miRNA_mRNA': '#circRNA_miRNA_mRNA'
            },
            axis=1
        )
        self.circ_events.submit_to_summary(total_count, type_='summary')

        category_count = self.res_df[['ev_id', 'category_1', 'category_2', 'category_3']].groupby(
            'ev_id'
        ).agg(
            'sum'
        ).pipe(
            self.circ_events.expand_to_all_events,
            fillna_value=0
        ).astype('int').rename(
            {
                'category_1': '#category_1',
                'category_2': '#category_2',
                'category_3': '#category_3',
            },
            axis=1
        )
        self.circ_events.submit_to_summary(category_count, type_='summary')

        if self.do_circRNA_RBP:
            circ_RBP_count = self.RBP_res_df[['ev_id', 'RBP']].drop_duplicates().rename(
                {
                    'RBP': '#circRNA_RBP'
                },
                axis=1
            ).groupby(
                'ev_id'
            ).agg(
                'count'
            ).pipe(
                self.circ_events.expand_to_all_events,
                fillna_value=0
            ).astype('int')
            self.circ_events.submit_to_summary(circ_RBP_count, type_='summary')

            if self.do_RBP_mRNA:
                circ_mRNA_via_RBP_count = self.RBP_res_df[['ev_id', 'target_gene']].drop_duplicates().rename(
                    {
                        'target_gene': '#circRNA_mRNA_via_RBP'
                    },
                    axis=1
                ).groupby(
                    'ev_id'
                ).agg(
                    'count'
                ).pipe(
                    self.circ_events.expand_to_all_events,
                    fillna_value=0
                ).astype('int')
                self.circ_events.submit_to_summary(circ_mRNA_via_RBP_count, type_='summary')

                circ_RBP_mRNA_count = self.RBP_res_df[['ev_id', 'RBP', 'target_gene']].drop_duplicates().assign(
                    circRNA_RBP_mRNA=1
                ).drop(
                    ['RBP', 'target_gene'],
                    axis=1
                ).groupby(
                    'ev_id'
                ).agg(
                    'count'
                ).pipe(
                    self.circ_events.expand_to_all_events,
                    fillna_value=0
                ).astype('int').rename(
                    {
                        'circRNA_RBP_mRNA': '#circRNA_RBP_mRNA'
                    },
                    axis=1
                )
                self.circ_events.submit_to_summary(circ_RBP_mRNA_count, type_='summary')

    def save_result(self, out_file):
        self.res_df.drop('ev_id', axis=1).to_csv(out_file, sep='\t', index=False)

    def save_RBP_result(self, out_file):
        self.RBP_res_df.drop('ev_id', axis=1).to_csv(out_file, sep='\t', index=False)

    def save_circRNAs_summary(self, out_file):
        self.circ_events.get_summary().to_csv(out_file, sep='\t', index=False)

    @staticmethod
    def _get_total_length(list_of_obj):
        return sum(map(len, list_of_obj))

    @classmethod
    def _get_uniq_exons(cls, anno_df):
        if anno_df.empty:
            uniq_exons_df = pd.DataFrame(
                [],
                columns=['exons', 'ev_id', 'total_len', 'exons_id']
            )
        else:
            uniq_exons_df = anno_df[['exons', 'ev_id']]\
                .drop_duplicates()\
                .reset_index(drop=True)

            uniq_exons_df['total_len'] = uniq_exons_df.apply(
                lambda s: cls._get_total_length(s['exons']),
                axis=1
            )

            uniq_exons_df['exons_id'] = uniq_exons_df.apply(
                lambda s: "exons_{}".format(s.name),
                axis=1
            )

        return uniq_exons_df

    @staticmethod
    def _get_category_df(res_df, to_binary=False):
        def get_category(AGO, MTB, MDB, ECR):
            if AGO:
                if MTB or ECR:
                    return '1'
                else:
                    return '2'
            else:
                if MTB or ECR:
                    return '2'
                else:
                    return '3'

        category_df = res_df.assign(
            AGO=res_df['num_AGO_supported_binding_sites'] > 0,
            MTB=res_df['miRTarBase'] == '1',
            MDB=res_df['miRDB'] == '1',
            ECR=res_df['ENCORI'] == '1'
        )[['AGO', 'MTB', 'MDB', 'ECR']].assign(
            category=lambda df: df.apply(lambda s: get_category(*s), axis=1)
        )[['category']]

        if to_binary:
            category_df = category_df.assign(
                category_1=(category_df['category'] == '1').apply(int),
                category_2=(category_df['category'] == '2').apply(int),
                category_3=(category_df['category'] == '3').apply(int)
            ).drop(
                'category',
                axis=1
            )

        return category_df


def get_mir_target_db(mir_tar_db_path):
    db = pd.read_csv(mir_tar_db_path, sep='\t', dtype='object')

    assert list(db.columns[:2]) == ['mirna', 'target_gene'], \
        ("The column names of the first two columns"
         " should be 'mirna' and 'target_gene'")

    return db


def debug_log(df, msg):
    logger.debug(msg)
    return df



#! /usr/bin/env python

import argparse
import subprocess as sp
import tempfile as tp
import io
import multiprocessing as mp
from collections import namedtuple
from collections.abc import Iterable
from operator import itemgetter


Bed6 = namedtuple('Bed6', ('chr_', 'start', 'end', 'name', 'score', 'strand'))


def get_fasta(bed_file, ref_file, bedtools_bin='bedtools'):
    cmd = [bedtools_bin, 'getfasta']
    cmd += ['-fi', ref_file]
    cmd += ['-bed', bed_file]
    cmd += ['-fo', '/dev/stdout']
    cmd += ['-s']
    # cmd += ['-name']

    res = sp.run(cmd, stdout=sp.PIPE, encoding='utf-8')

    return res.stdout


def blastn(query, subject, blastn_bin='blastn'):
    cmd = [blastn_bin]
    cmd += ['-task', 'blastn']
    cmd += ['-query', query]
    cmd += ['-subject', subject]
    cmd += ['-outfmt', '6']
    cmd += ['-strand', 'minus']
    # cmd += ['-word_size', '11']

    res = sp.run(cmd, stdout=sp.PIPE, encoding='utf-8')

    return res.stdout


def save_tmp(data, dir_='.'):
    tmp_file = tp.NamedTemporaryFile(dir=dir_)

    with open(tmp_file.name, 'w') as tmp_out:
        if isinstance(data, str):
            tmp_out.write(data)

        elif isinstance(data, Iterable):
            print(*data, sep='\t', file=tmp_out)

    return tmp_file


TITLES = [
    'circRNA_id',
    'region1',
    'region2',
    'matches',
    'alignment_len',
    'mismatch',
    'gapopen',
    'alignment_start_1',
    'alignment_end_1',
    'alignment_start_2',
    'alignment_end_2',
    'E_value',
    'bit_score',
    'region1_type',
    'region2_type'
]


RCS_data = namedtuple('RCS_data', TITLES[1:13])


def read_blastn_result(blastn_result):
    for line in blastn_result.rstrip('\n').split('\n'):
        data = line.rstrip('\n').split('\t')
        data = RCS_data(*data)
        yield data


def is_disjoint(r1, r2):
    r1, r2 = sorted([r1, r2], key=itemgetter(0))

    if r1[1] < (r2[0] - 1):
        return True
    else:
        return False


def correct_results(blastn_result):
    all_alignments = []

    read_data = read_blastn_result(blastn_result)

    with io.StringIO() as out:
        for data in read_data:
            region1 = (int(data.alignment_start_1), int(data.alignment_end_1))
            region2 = (int(data.alignment_end_2), int(data.alignment_start_2))

            if is_disjoint(region1, region2):

                # check duplicates
                if (region2, region1) not in all_alignments:
                    all_alignments.append((region1, region2))

                    print(*data, sep='\t', file=out)

        return out.getvalue()


def read_circ_file(circ_file):
    for line in circ_file:
        data = line.rstrip('\n').split('\t')
        chr_ = data[0]
        pos1, pos2 = sorted(map(int, data[1:3]))
        strand = data[3]

        if len(data) >= 5:
            circ_id = data[4]
        else:
            if strand == '+':
                circ_id = f"{chr_}:{pos2}|{pos1}({strand})"
            elif strand == '-':
                circ_id = f"{chr_}:{pos1}|{pos2}({strand})"

        if strand == '+':
            up_type = 'acceptor'
            down_type = 'donor'
        elif strand == '-':
            up_type = 'donor'
            down_type = 'acceptor'

        yield (chr_,
               pos1,
               pos2,
               strand,
               circ_id,
               up_type,
               down_type)


def get_RCS_output(circ_id, blastn_result, region1_type, region2_type):
    with io.StringIO() as tmp_out:
        if blastn_result:
            for res_line in blastn_result.rstrip('\n').split('\n'):
                print(circ_id, res_line, region1_type, region2_type, sep='\t', file=tmp_out)
        else:
            print(circ_id, *([''] * 12), region1_type, region2_type, sep='\t', file=tmp_out)

        return tmp_out.getvalue()


class RCS:
    def __init__(self, ref_file, dist, tmp_dir='.'):
        self.ref_file = ref_file
        self.dist = dist
        self.tmp_dir = tmp_dir

    def get_RCS(self, chr_, pos1, pos2, strand, circ_id, up_type, down_type):
        # upstream
        up_bed = Bed6(chr_, max(pos1 - self.dist, 0), pos1, circ_id, '.', strand)
        up_bed_file = save_tmp(up_bed, dir_=self.tmp_dir)

        up_fa = get_fasta(up_bed_file.name, self.ref_file)
        up_fa_file = save_tmp(up_fa, dir_=self.tmp_dir)

        # downstream
        down_bed = Bed6(chr_, pos2 - 1, pos2 - 1 + self.dist, circ_id, '.', strand)
        down_bed_file = save_tmp(down_bed, dir_=self.tmp_dir)

        down_fa = get_fasta(down_bed_file.name, self.ref_file)
        down_fa_file = save_tmp(down_fa, dir_=self.tmp_dir)

        # cross
        if strand == '+':
            blastn_result = blastn(down_fa_file.name, up_fa_file.name)
            RCS_result = get_RCS_output(circ_id, blastn_result, down_type, up_type)
        elif strand == '-':
            blastn_result = blastn(up_fa_file.name, down_fa_file.name)
            RCS_result = get_RCS_output(circ_id, blastn_result, up_type, down_type)

        # within upstream region
        blastn_result_up = blastn(up_fa_file.name, up_fa_file.name)
        if blastn_result_up:
            blastn_result_up = correct_results(blastn_result_up)
        RCS_result_up = get_RCS_output(circ_id, blastn_result_up, up_type, up_type)

        # within downstream region
        blastn_result_down = blastn(down_fa_file.name, down_fa_file.name)
        if blastn_result_down:
            blastn_result_down = correct_results(blastn_result_down)
        RCS_result_down = get_RCS_output(circ_id, blastn_result_down, down_type, down_type)

        return RCS_result, RCS_result_up, RCS_result_down


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--dist', default=2000, type=int)
    parser.add_argument('ref_file')
    parser.add_argument('circ_file', type=argparse.FileType('r'), help="4-columns TSV file: (chr, pos1, pos2, strand)")
    parser.add_argument('-p', dest='num_proc', type=int, default=1)

    return parser


def cli():
    parser = create_parser()
    args = parser.parse_args()

    print(*TITLES, sep='\t')

    with tp.TemporaryDirectory(prefix='get_RCS.tmp.', dir='.') as tmp_dir:
        rcs = RCS(args.ref_file, args.dist, tmp_dir)

        with mp.Pool(processes=args.num_proc) as pool:
            RCS_results = [
                pool.apply_async(
                    rcs.get_RCS,
                    circ_data
                )
                for circ_data in read_circ_file(args.circ_file)
            ]

            for result in RCS_results:
                RCS_result, RCS_result_up, RCS_result_down = result.get()

                print(RCS_result, end='')
                print(RCS_result_up, end='')
                print(RCS_result_down, end='')


if __name__ == "__main__":
    cli()



import os.path
import re
import pandas as pd
import numpy as np
from collections import Counter, defaultdict
from circmimi.annotation import Annotator
from circmimi.ambiguous import AmbiguousChecker


class CircEvents:
    INPUT_COLUMNS = ('chr', 'pos1', 'pos2', 'strand', 'circ_id')
    DONOR_ACCEPTOR_COLUMNS = ('chr', 'donor', 'acceptor', 'strand', 'circ_id')

    def __init__(self, filename):
        self._filename = filename
        self.original_df = self._read_file(self._filename)
        self.df = self._get_donor_acceptor_df(self.original_df)

        self._summary_columns = {
            'summary': [],
            'filters': []
        }

    def _read_file(self, filename):
        if not os.path.exists(filename):
            raise FileNotFoundError(f"The file \"{filename}\" does not exist!")

        circRNA_data = get_circRNA_data(open(filename).read())

        if not circRNA_data:
            raise NoValidCircRNAEvents("Empty!")

        all_num_cols = sorted(set([len(circRNA) for circRNA in circRNA_data]))

        if not ((len(all_num_cols) == 1) and (all_num_cols[0] in [4, 5])):
            raise NoValidCircRNAEvents("Format error!")

        if all_num_cols[0] == 5:
            self.circ_ids_specified = True
        else:
            self.circ_ids_specified = False

        df = pd.read_csv(
            filename,
            sep='\t',
            names=self.INPUT_COLUMNS,
            dtype={
                'chr': 'category',
                'pos1': 'int',
                'pos2': 'int',
                'strand': 'category'
            }
        ).rename_axis('ev_id')

        return df

    @classmethod
    def _get_donor_acceptor(cls, s):
        chr_, pos1, pos2, strand, circ_id = s[list(cls.INPUT_COLUMNS)]
        pos1, pos2 = sorted([pos1, pos2])

        if strand == '+':
            donor_site = pos2
            acceptor_site = pos1
        elif strand == '-':
            donor_site = pos1
            acceptor_site = pos2

        res = [chr_, donor_site, acceptor_site, strand, circ_id]
        res = pd.Series(res, index=cls.DONOR_ACCEPTOR_COLUMNS)

        return res

    @classmethod
    def _get_donor_acceptor_df(cls, original_df):
        if original_df.empty:
            df = pd.DataFrame([], columns=cls.DONOR_ACCEPTOR_COLUMNS)
        else:
            df = original_df.apply(cls._get_donor_acceptor, axis=1)

        return df

    def expand_to_all_events(self, ev_df, fillna_value):
        expanded_df = self.original_df.reset_index()[['ev_id']].merge(
            ev_df,
            on='ev_id',
            how='left'
        ).fillna(fillna_value).set_index('ev_id')

        return expanded_df

    def submit_to_summary(self, summary_column, type_):
        self._summary_columns[type_].append(summary_column)

    @staticmethod
    def _get_host_genes(anno_df):

        def get_gene_symbol(anno_df):
            return anno_df['transcript'].apply(
                lambda t: t.gene.gene_symbol
            )

        host_gene_df = anno_df.assign(
            host_gene=get_gene_symbol
        )[[
            'ev_id',
            'host_gene'
        ]].drop_duplicates(
        ).groupby(
            'ev_id'
        ).agg(lambda genes: ','.join(sorted(genes))).reset_index()

        return host_gene_df

    @staticmethod
    def _get_circ_ids(host_genes):
        genes = host_genes['host_gene']
        gene_count = Counter(genes)
        idx_dict = defaultdict(int)

        circ_ids = []
        for gene in genes:
            if gene_count[gene] == 1:
                circ_ids.append(f'circ{gene}')
            else:
                idx_dict[gene] += 1
                circ_ids.append(f'circ{gene}_{idx_dict[gene]}')

        circ_ids_df = host_genes.assign(
            circ_id=pd.Series(circ_ids)
        )[['ev_id', 'circ_id']]

        return circ_ids_df

    def _update_circ_ids(self, circ_ids):
        all_ev_with_circ_id = self.expand_to_all_events(circ_ids, np.nan)
        self.original_df = self.original_df.assign(circ_id=all_ev_with_circ_id)
        self.df = self.df.assign(circ_id=all_ev_with_circ_id)

    def _append_host_genes(self, host_genes):
        all_ev_with_host_gene = self.expand_to_all_events(host_genes, np.nan)
        self.original_df = self.original_df.assign(host_gene=all_ev_with_host_gene)
        self.df = self.df.assign(host_gene=all_ev_with_host_gene)

    def check_annotation(self, anno_db_file):
        self._annotator = Annotator(anno_db_file)
        self.anno_df, anno_status = self.df.pipe(self._annotator.annotate)

        self._host_genes = self._get_host_genes(self.anno_df)
        self._append_host_genes(self._host_genes)

        if not self.circ_ids_specified:
            self._circ_ids = self._get_circ_ids(self._host_genes)
            self._update_circ_ids(self._circ_ids)

        self.submit_to_summary(anno_status, type_='filters')

    def check_ambiguous(self,
                        anno_db_file,
                        ref_file,
                        other_ref_file,
                        work_dir='.',
                        num_proc=1):

        self.checker = AmbiguousChecker(
            anno_db_file,
            ref_file,
            other_ref_file,
            work_dir=work_dir,
            num_proc=num_proc
        )
        checking_result = self.checker.check(self.df)

        self.submit_to_summary(checking_result, type_='filters')

    @staticmethod
    def _merge_columns(df, column_dfs):
        merged_df = df

        for col_df in column_dfs:
            merged_df = merged_df.merge(col_df, on='ev_id', how='left')

        return merged_df

    def get_filters_results(self, show_detail=False):
        filters_df = pd.DataFrame(self.df.index).pipe(
            self._merge_columns,
            self._summary_columns['filters']
        )

        if show_detail:
            filters_df = self.original_df.pipe(
                self._merge_columns,
                [filters_df]
            ).set_index('ev_id')

        return filters_df

    def get_summary(self):
        filters_df = self.get_filters_results()

        pass_column = filters_df.fillna('2').set_index(
            'ev_id'
        ).astype(
            int
        ).agg(
            'sum',
            axis=1
        ).eq(
            3
        ).replace(
            {
                True: 'yes',
                False: 'no'
            }
        ).reset_index().rename({0: 'pass'}, axis=1)

        summary_df = self.original_df.reset_index().pipe(
            self._merge_columns,
            self._summary_columns['summary'] + [pass_column, filters_df.fillna('NA')]
        ).set_index('ev_id')

        return summary_df

    @property
    def _passed_events(self):
        pass_df = self.get_summary()[lambda df: df['pass'] == 'yes'].reset_index()[['ev_id']]
        return pass_df

    @property
    def clear_df(self):
        return self.original_df.merge(self._passed_events, on='ev_id').set_index('ev_id')

    @property
    def clear_anno_df(self):
        return self.anno_df.merge(self._passed_events, on='ev_id', how='inner')

    @property
    def region_id(self):
        region_id_df = self.df.apply(
            lambda s: f"{s['chr']}:{s['donor']}|{s['acceptor']}({s['strand']})",
            axis=1
        ).rename('region_id')

        return region_id_df


class Formatter:
    @staticmethod
    def format_input_data(input_data):
        '''Deal with spaces between rows.'''
        input_data = re.sub(r"\s*\n\s*", "\n", input_data).strip('\n')
        return input_data

    @staticmethod
    def format_circRNA_event(circRNA_text):
        '''Deal with spaces between columns.'''
        circRNA_text = re.sub(r"\s+", "\t", circRNA_text).strip('\t')
        return circRNA_text

    @staticmethod
    def format_chromosome_name(chr_name):
        chr_name = chr_name.lstrip('chr')

        if chr_name == 'MT':
            chr_name = 'M'

        return chr_name


def get_circRNA_data(input_data):
    all_circRNA_data = []

    input_data = Formatter.format_input_data(input_data)
    for circRNA_text in input_data.split('\n'):
        circRNA_text = Formatter.format_circRNA_event(circRNA_text)

        if circRNA_text == '':
            continue

        circRNA_data = circRNA_text.split('\t')
        # circRNA_data[0] = Formatter.format_chromosome_name(circRNA_data[0])

        all_circRNA_data.append(circRNA_data)

    return all_circRNA_data


class NoValidCircRNAEvents(Exception):
    pass





#! /usr/bin/env python
import click
import os
import logging


logger = logging.getLogger(__name__)


@click.group()
@click.version_option()
@click.option('--debug', 'debug_mode', is_flag=True)
def cli(debug_mode):
    """
    A toolset for investigating the interactions between circRNA, miRNA, and mRNA.
    """

    root_logger = logging.getLogger()

    ch = logging.StreamHandler()

    if debug_mode:
        logger.warning('Under debug mode')
        root_logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s (%(name)s) [%(levelname)s] %(message)s')
    else:
        root_logger.setLevel(logging.INFO)
        formatter = logging.Formatter('%(asctime)s - %(message)s')

    ch.setFormatter(formatter)
    root_logger.addHandler(ch)


@cli.command('interactions')
@click.option('-r', '--ref', 'ref_dir', type=click.Path(), metavar="REF_DIR", required=True)
@click.option('-i', '--circ', 'circ_file', metavar="CIRC_FILE", required=True)
@click.option('-o', '--out-prefix', 'out_prefix', default='./', metavar="OUT_PREFIX")
@click.option('-p', '--num_proc', default=1, type=click.INT, metavar="NUM_PROC",
    help="Number of processes")
@click.option('--checkAA', 'checkAA', is_flag=True,
    help="Check if the circRNA has ambiguous alignments.", hidden=True)
@click.option('--no-pvalue-filtering', 'pv_filter', flag_value=False, default=True,
    help="If this option is set, the results will contain all interactions without P-values filtering.")
@click.option('--miranda-sc', 'sc', metavar='S', type=click.FLOAT, default=155, help='(Default: 155)')
@click.option('--miranda-en', 'en', metavar='-E', type=click.FLOAT, default=-20, help='(Default: -20)')
@click.option('--miranda-scale', 'scale', metavar='Z', type=click.FLOAT)
@click.option('--miranda-strict', 'strict', is_flag=True)
@click.option('--miranda-go', 'go', metavar='-X', type=click.FLOAT)
@click.option('--miranda-ge', 'ge', metavar='-Y', type=click.FLOAT)
def predict_interactions(circ_file,
                         ref_dir,
                         out_prefix,
                         num_proc,
                         checkAA,
                         pv_filter,
                         **miranda_options):

    """
    Predict the interactions.

    Predict the interactions between circRNA-miRNA-mRNA.
    """

    logger.info('Preparing ...')

    from circmimi.utils import add_prefix

    output_dir = os.path.dirname(out_prefix)
    if output_dir == '':
        output_dir = '.'

    if output_dir != '.':
        os.makedirs(output_dir, exist_ok=True)

    from circmimi.reference.config import get_refs
    anno_db, ref_file, mir_ref, mir_target, other_transcripts, AGO_data, RBP_data, RBP_target = get_refs(ref_dir)

    if checkAA:
        other_ref_file = other_transcripts
    else:
        other_ref_file = None

    miranda_options_list = []
    for k, v in miranda_options.items():
        if v is not None:
            cmd = ['-' + k]
            if k != 'strict':
                cmd.append(str(v))

            miranda_options_list.extend(cmd)

    from circmimi.circmimi import Circmimi
    circmimi_result = Circmimi(
        anno_db,
        ref_file,
        mir_ref,
        mir_target,
        AGO_data,
        RBP_data,
        RBP_target,
        other_ref_file,
        work_dir=output_dir,
        num_proc=num_proc,
        pv_filter=pv_filter,
        miranda_options=miranda_options_list
    )

    logger.info('Starting the main pipeline.')
    circmimi_result.run(circ_file)
    logger.info('Pipeline completed.')

    logger.info('Saving results ...')
    res_file = add_prefix('all_interactions.miRNA.tsv', out_prefix)
    circmimi_result.save_result(res_file)

    circ_target_pv_file = add_prefix('circRNA_target_gene.pvalue.tsv', out_prefix)
    circmimi_result.circ_target_df_with_pv.to_csv(circ_target_pv_file, sep='\t', index=False)

    logger.info('miRNA part ... done')

    # RBP_res_file = add_prefix('all_interactions.RBP.tsv', out_prefix)
    # circmimi_result.save_RBP_result(RBP_res_file)
    # logger.info('RBP part ... done')

    summary_file = add_prefix('summary_list.tsv', out_prefix)
    circmimi_result.save_circRNAs_summary(summary_file)
    logger.info('summary file ... done')
    logger.info('All results are saved.')

    logger.info('Process completed.')


@cli.command('genref')
@click.option('--species', 'species', metavar="SPECIES_KEY", required=True, help="eg. hsa, mmu, rno, ...")
@click.option('--source', 'source', metavar="SOURCE", required=True,
    help="""
        Available sources are "gencode", "ensembl", "ensembl_plants", and "ensembl_metazoa"
    """)
@click.option('--gencode', 'source', flag_value='gencode', type=click.STRING, help="--source gencode", hidden=True)
@click.option('--ensembl', 'source', flag_value='ensembl', type=click.STRING, help="--source ensembl", hidden=True)
@click.option('--version', 'version', default='current', metavar="VERSION",
    help="""
        The release version. If not specified, it will be automatically set to the latest version of the SOURCE.
    """)
@click.option('--init', 'init', is_flag=True, help="Create an init template ref_dir.", hidden=True)
@click.argument('ref_dir')
def generate_references(species, source, version, ref_dir, init):
    """
    Generate the references.                                          

    This command is used to generate the needed references for CircMiMi.
    It would download the genome ane annotation files automatically, with user-specified version.

    The generated reference files will be saved in the REF_DIR.
    """

    os.makedirs(ref_dir, exist_ok=True)

    from circmimi.reference.config import RefConfig
    config = RefConfig()

    if not init:
        from circmimi.reference import genref
        info, ref_files = genref.generate(species, source, version, ref_dir)

        config['info'].update(info)
        config['refs'].update(ref_files)

    config.write(ref_dir)


@cli.command('gendb', hidden=True)
@click.argument('gtf_path')
@click.argument('db_path', metavar='OUT_PATH')
def generate_annotation_database(gtf_path, db_path):
    from circmimi.reference import gendb

    gendb.generate(gtf_path, db_path)


@cli.command('genmirdb', hidden=True)
@click.option('--species', 'species', metavar="SPECIES_KEY", required=True)
@click.option('--version', 'version', default='current', metavar="VERSION", required=True)
@click.option('-r', '--ref', 'ref_dir', type=click.Path(), metavar="REF_DIR", required=True)
@click.option('-o', '--out_file', 'out_file', metavar="OUT_FILE", required=True)
@click.option('-a', '--show-accession', 'show_accession', is_flag=True)
def generate_miRNA_database(species, version, ref_dir, out_file, show_accession):
    os.makedirs(ref_dir, exist_ok=True)

    from circmimi.reference import genmirdb

    genmirdb.generate(species, version, ref_dir, out_file, show_accession)


@cli.group(hidden=True)
def check():
    pass


@check.command('annotation')
@click.argument('anno_db')
@click.argument('circ_file')
@click.argument('out_file')
def check_annotation(circ_file, anno_db, out_file):
    from circmimi.circ import CircEvents

    circ_events = CircEvents(circ_file)
    circ_events.check_annotation(anno_db)

    circ_events.get_filters_results(True).to_csv(out_file, sep='\t', index=False)


@check.command('ambiguous')
@click.argument('ref_file')
@click.argument('other_ref_file')
@click.argument('circ_file')
@click.argument('out_file')
@click.option('-p', '--num_proc', default=1, type=click.INT,
              metavar="NUM_PROC", help="Number of processes")
def check_ambiguous(ref_file, other_ref_file, circ_file, out_file, num_proc):
    import subprocess as sp
    import tempfile as tp

    tmp_fa = tp.NamedTemporaryFile(
        dir='.',
        prefix='flanking_seq.',
        suffix='.fa'
    )

    sp.run(['get_flanking_seq.py', ref_file, circ_file, tmp_fa.name])
    sp.run(['checkAA_reads.py', '-rg', ref_file, '-ro', other_ref_file,
            tmp_fa.name, out_file, '-p', str(num_proc)])


@check.command('RCS')
@click.argument('ref_file')
@click.argument('circ_file')
@click.argument('out_file')
@click.option('-d', '--dist', default=10000, type=click.INT)
@click.option('-m', '--min-matches', default=80, type=click.FLOAT)
@click.option('-l', '--min-aln-len', default=50, type=click.INT)
@click.option('-b', '--min-bitscore', default=100, type=click.FLOAT)
@click.option('-p', '--num_proc', default=1, type=click.INT,
              metavar="NUM_PROC", help="Number of processes")
def check_RCS(ref_file,
              circ_file,
              out_file,
              dist,
              min_matches,
              min_aln_len,
              min_bitscore,
              num_proc):

    import subprocess as sp

    cmd1 = ['get_RCS.py', ref_file, circ_file, '--dist', dist, '-p', num_proc]
    cmd2 = ['RCS_filter.py', '-', '-m', min_matches, '-l', min_aln_len, '-b', min_bitscore]
    cmd3 = ['get_RCS_summary.py', '--circ_file', circ_file, '-']

    cmd1 = [str(c) for c in cmd1]
    cmd2 = [str(c) for c in cmd2]
    cmd3 = [str(c) for c in cmd3]

    with open(out_file, 'w') as out:
        p1 = sp.Popen(cmd1, stdout=sp.PIPE, encoding='UTF-8')
        p2 = sp.Popen(cmd2, stdin=p1.stdout, stdout=sp.PIPE, encoding='UTF-8')
        p3 = sp.Popen(cmd3, stdin=p2.stdout, stdout=out, encoding='UTF-8')

        p1.wait()
        p2.wait()
        p3.wait()


@cli.command('checking')
@click.option('-i', '--circ', 'circ_file', metavar="CIRC_FILE", required=True)
@click.option('-r', '--ref', 'ref_dir', type=click.Path(), metavar="REF_DIR",
              required=True)
@click.option('-o', '--out-prefix', 'out_prefix', default='./',
              metavar="OUT_PREFIX")
@click.option('-p', '--num_proc', default=1, type=click.INT,
              metavar="NUM_PROC", help="Number of processes")
@click.option('-d', '--dist', default=10000, type=click.INT)
@click.pass_context
def default_checking(ctx, circ_file, ref_dir, out_prefix, num_proc, dist):
    """
    Check the circRNAs.

    Check the circRNAs if there are
    (i) ambiguous alignments for the fragment around the circRNA junction;
    (ii) RCS pair in the flanking sequences.
    """

    import tempfile as tp
    from circmimi.reference.config import get_refs
    from circmimi.utils import add_prefix

    output_dir = os.path.dirname(out_prefix)
    if output_dir == '':
        output_dir = '.'

    if output_dir != '.':
        os.makedirs(output_dir, exist_ok=True)

    anno_db, ref_file, _, _, other_transcripts, _, _, _ = get_refs(ref_dir)

    annotation_result_file = tp.NamedTemporaryFile(
        dir=output_dir,
        prefix='check.annotation.',
        suffix='.tsv'
    )

    ambiguous_result_file = tp.NamedTemporaryFile(
        dir=output_dir,
        prefix='check.AA.',
        suffix='.tsv'
    )
    RCS_result_file = tp.NamedTemporaryFile(
        dir=output_dir,
        prefix='check.RCS.',
        suffix='.tsv'
    )
    out_file = add_prefix('checking.results.tsv', out_prefix)

    ctx.invoke(
        check_annotation,
        circ_file=circ_file,
        anno_db=anno_db,
        out_file=annotation_result_file.name
    )

    ctx.invoke(
        check_ambiguous,
        ref_file=ref_file,
        other_ref_file=other_transcripts,
        circ_file=circ_file,
        out_file=ambiguous_result_file.name,
        num_proc=num_proc
    )

    ctx.invoke(
        check_RCS,
        ref_file=ref_file,
        circ_file=circ_file,
        out_file=RCS_result_file.name,
        num_proc=num_proc,
        dist=dist
    )

    import pandas as pd

    ann_df = pd.read_csv(annotation_result_file.name, sep='\t', dtype='object')
    amb_df = pd.read_csv(ambiguous_result_file.name, sep='\t', dtype='object')
    rcs_df = pd.read_csv(RCS_result_file.name, sep='\t', dtype='object')
    result_df = pd.concat(
        [
            ann_df,
            amb_df.drop('circRNA_id', axis=1),
            rcs_df.drop('circRNA_id', axis=1)
        ],
        axis=1
    )
    result_df.to_csv(out_file, sep='\t', index=False)


@cli.group(hidden=True)
def network():
    """
    Command to create the network file for Cytoscape.
    """

    pass


@network.command('create')
@click.argument('in_file')
@click.argument('out_file')
@click.option('-1', 'idx_circRNA', type=int, default=1,
              help='column key for circRNAs.')
@click.option('-2', 'idx_mediator', type=int, default=2,
              help='column key for mediators.')
@click.option('-3', 'idx_mRNA', type=int, default=3,
              help='column key for mRNAs.')
@click.option('-f', '--format', 'format_', default='xgmml',
              help="Assign the format of the OUT_FILE.", hidden=True)
def create_network(in_file, out_file, idx_circRNA, idx_mediator, idx_mRNA, format_):
    """
    Create the network file.

    This command would transform the IN_FILE to XGMML format,
    and save to OUT_FILE.
    """

    from circmimi.network.network import CyNetwork, Layout, Style

    network = CyNetwork()
    network.load_data(
        in_file,
        k1=idx_circRNA,
        k2=idx_mediator,
        k3=idx_mRNA
    )

    layout = Layout()
    style = Style()

    network.apply_layout(layout)
    network.apply_style(style)

    if format_ == "xgmml":
        network.to_xgmml(out_file)


@cli.command('visualize')
@click.argument('in_file')
@click.argument('out_file')
@click.option('-1', 'idx_circRNA', type=int, default=5,
              help='column key for circRNAs.')
@click.option('-2', 'idx_mediator', type=int, default=7,
              help='column key for mediators.')
@click.option('-3', 'idx_mRNA', type=int, default=13,
              help='column key for mRNAs.')
@click.option('-f', '--format', 'format_', default='xgmml',
              help="Assign the format of the OUT_FILE.", hidden=True)
@click.option('--header', 'header', flag_value=True, type=click.BOOL,
              default=True, hidden=True,
              help="Use this option to specify if there are titles in the IN_FILE.")
@click.option('--no-header', 'header', flag_value=False, type=click.BOOL, default=True,
              help="Use this option to specify if there are no titles in the IN_FILE.")
def visualize_interactions(in_file,
                           out_file,
                           idx_circRNA,
                           idx_mediator,
                           idx_mRNA,
                           header,
                           format_):

    """
    Visualize the interactions.

    This command is used to convert the file of interactions into XGMML format,
     which then can be loaded into Cytoscape for visualization and further analysis.
    """

    from circmimi.network.network import CyNetwork, Layout, Style

    network = CyNetwork()
    network.load_data(
        in_file,
        k1=idx_circRNA,
        k2=idx_mediator,
        k3=idx_mRNA,
        header=header
    )

    layout = Layout()
    style = Style()

    network.apply_layout(layout)
    network.apply_style(style)

    if format_ == "xgmml":
        network.to_xgmml(out_file)


@cli.group('update_mirna', hidden=True)
def update_mirna():
    pass


@update_mirna.command('genmaps')
@click.option('--from', 'from_', required=True)
@click.option('--to', 'to_', required=True)
@click.option('--species')
@click.option('-o', '--out-prefix', 'out_prefix', default='./', metavar="OUT_PREFIX")
def generate_miRNAs_mappings(from_, to_, species, out_prefix):
    from circmimi.reference.mirbase import MatureMiRNAUpdater
    from circmimi.utils import add_prefix

    output_dir, prefix_name = os.path.split(out_prefix)

    if output_dir == '':
        output_dir = '.'

    if output_dir != '.':
        os.makedirs(output_dir, exist_ok=True)

    updater = MatureMiRNAUpdater(from_, to_, species)
    updater.create(output_dir)

    out_file = 'miRNA.maps_{}_to_{}.tsv'.format(from_, to_)
    out_file = add_prefix(out_file, prefix_name)
    out_file = os.path.join(output_dir, out_file)

    updater.save(out_file)


@update_mirna.command('update')
@click.argument('in_file')
@click.argument('out_file')
@click.option('-m', '--maps', 'mapping_file', metavar="MAPPING_FILE", required=True)
@click.option('-k', '--col-key', 'column_key', type=click.INT, default=1,
              help="The column number of miRNA IDs.")
@click.option('-i', '--inplace', is_flag=True)
@click.option('-R', '--remove-deleted', is_flag=True)
def update_miRNAs(in_file, out_file, mapping_file, column_key, inplace, remove_deleted):
    column_key = column_key - 1

    from circmimi.reference.mirbase import MatureMiRNAUpdater
    updater = MatureMiRNAUpdater(None, None, None)
    updater.load_maps(mapping_file)
    updater.update_file(in_file, out_file, column_key, inplace, remove_deleted)


@cli.command('calc-pvalue', hidden=True)
@click.argument('interaction_file')
@click.option('--mir_ref_file')
@click.option('--mir_target_file')
@click.option('-o', '--out_prefix', default='./')
def calculate_pvalue_of_interactions(interaction_file, out_prefix, mir_ref_file, mir_target_file):
    """
    This command is used to calculate the P-value of the interaction between circRNA and target gene.
    """

    import pandas as pd
    from circmimi.stats import do_the_hypergeometric_test
    from circmimi.circmimi import get_mir_target_db

    df = pd.read_csv(interaction_file, sep='\t', dtype='object')
    circ_mir_target_df = df[['circ_id', 'mirna', 'target_gene']].drop_duplicates().reset_index(drop=True)

    mir_target_db = get_mir_target_db(mir_target_file)

    circ_target_df_with_pv = do_the_hypergeometric_test(
        circ_mir_target_df,
        mir_ref_file,
        mir_target_db
    )

    circ_target_df_with_pv.to_csv(out_prefix + "circRNA_target_gene.pvalue.tsv", sep='\t', index=False)

    result_df = df.merge(
        circ_target_df_with_pv[[
            'circ_id',
            'target_gene',
            'p_value',
            'bh_corrected_p_value',
            'bonferroni_corrected_p_values',
        ]],
        on=['circ_id', 'target_gene'],
        how='left'
    )

    result_df.to_csv(out_prefix + "all_interactions.miRNA.pvalue.tsv", sep='\t', index=False)


if __name__ == "__main__":
    cli()
